from collections import namedtuple

# These namedtuples are lightweight readable containers for Layer/Attribute info
Layer = namedtuple('Layer', ['name', 'tess_id'])
Attribute = namedtuple('Attribute', ['name', 'unit'])



"""
GeoTessGrid Python definitions.

"""
import numpy as np

import geotess.libgeotess as lib

class Grid(object):
    """
    Manages the geometry and topology of a multi-level triangular tessellation
    of a unit sphere. It knows:

    * the positions of all the vertices,
    * the connectivity information that defines how vertices are connected to
      form triangles,
    * for each triangle it knows the indexes of the 3 neighboring triangles,
    * for each triangle it knows the index of the triangle which is a descendant
      at the next higher tessellation level, if there is one.
    * information about which triangles reside on which tessellation level 

    """
    def __init__(self, gridfile=None):
        """
        Construct a grid from a file.

        Parameters
        ----------
        gridfile : str
            Full path to GeoTess grid file.  None results in an empty Grid
            instance.

        Attributes
        ----------
        _grid : geotess.GeoTessGrid
            Low-level access to the GeoTessGrid instance.
        tesselations
        levels
        triangles
        vertices

        """
        if gridfile:
            self._grid = lib.GeoTessGrid()
            self._grid.loadGrid(gridfile)
        else:
            self._grid = None

    @classmethod
    def from_geotessgrid(cls, gtgrid):
        """
        Constructor that wraps a geotess.libgeotess.GeoTessGrid instance.

        """
        # TODO: rename this to "wrap"
        g = cls()
        g._grid = gtgrid

        return g

    def triangles(self, tess, level, masked=False):
        """
        Get tessellation triangles, as integer indices into the corresponding
        array of vertices.

        Use these "triangles" (vertex indices) to index into the corresponding
        Model.vertices.

        Parameters
        ----------
        tess : int
            The integer index of the target tessellation.
        level : int
            The integer of the target tessellation level.
        masked : bool
            If False, only return un-masked triangles.  Otherwise, return all.
            Not yet implemented.

        Returns
        -------
        triangles : numpy.ndarray of ints (Ntriangles x 3)
            Each row contains (unordered?) integer indices into the
            corresponding vertex array, producing the triangle coordinates.

        See Also
        --------
        Model.vertices

        """
        # XXX: handle None values from a grid.triangles() call.
        #   it should forward to an exposed _grid.getTriangles() method.

        # get the integer ids of all the triangles in this layer and level
        first_triangle_id = self._grid.getFirstTriangle(tess, level)
        last_triangle_id = self._grid.getLastTriangle(tess, level)
        triangle_ids = range(first_triangle_id, last_triangle_id + 1)

        # get the vertex indices of all the triangles as an iteger array
        triangles = np.empty((len(triangle_ids), 3), dtype=np.int)
        for i, triangle_id in enumerate(triangle_ids):
            triangles[i,:] = self._grid.getTriangleVertexIndexes(triangle_id)

        return triangles

    def vertices(self, tess=None, level=None, masked=False):
        """
        Get geocentric coordinates of tessellation vertices.

        Parameters
        ----------
        layer : str or int
            The string name or integer index of the target layer.
            Not yet implemented.
        level : int
            The integer of the target tessellation level.
            Not yet implemented.
        masked : bool
            If False, only return un-masked vertices.  Otherwise, return all.
            Not yet implemented.

        Returns
        -------
        vertices : numpy.ndarray of floats (Nvert X 2)
            Geographic coordinates of vertices.

        """
        # XXX: implement tess, level keywords.
        gt_util = lib.GeoTessUtils()
        unit_vertices = self._grid.getVertices()
        geo_vertices = np.empty((unit_vertices.shape[0], 2), dtype=np.float)
        for unit_vertex, geo_vertex in zip(unit_vertices, geo_vertices):
            # this syntax looks funny, because we never put geo_vertex back into
            # the original array, but it works b/c iterating through rows
            # in an array produces references.  so, manipulating these row
            # references is actually manipulating the main array.
            geo_vertex[0] = gt_util.getLonDegrees(unit_vertex)
            geo_vertex[1] = gt_util.getLatDegrees(unit_vertex)

        return geo_vertices


    def __str__(self):
        return str(self._grid.toString())



#import os
#import sys
#
#sys.path.append(os.path.dirname(__file__) + os.path.sep + 'lib' + os.path.sep + 'geotess.jar')

__version__ = '0.1.1'

from geotess.utils import Layer, Attribute
from geotess.grid import Grid
from geotess.model import Model



"""
Home of the GeoTess Model.

Model is the main user-facing GeoTess class.  GeoTess Grids are build by 
geotessbuilder, and rarely (never) directly manipulated.  GeoTess MetaData
instances are mostly hidden within GeoTess Models.  The Model is the access
layer and manager of each of these other classes.

"""
from datetime import datetime

import numpy as np

from geotess import __version__
from geotess.utils import Layer, Attribute
from geotess.grid import Grid
import geotess.libgeotess as lib

def _geotess_layer_names(name_list):
    """ Convert Python list of names to GeoTess-style name list.
    """
    # provide a list of string layer names
    # converts to uppercase, spaces to underscores, and joins them with "; "
    layer_names = '; '.join([name.upper().replace(' ', '_') for name in name_list])

    return layer_names

def _attributes_from_strings(gt_names, gt_units):
    """Produce geotess.Attribute tuples from GeoTess-style names and units lists.
    """
    names = gt_names.split(';')
    units = gt_units.split(';')

    return [Attribute(name=name, unit=unit) for name, unit in zip(names, units)]

def _geotess_attributes(attrib_tuples):
    """
    Convert geotess.Attribute tuples to GeoTess-style attribute name list and
    attribute unit list.

    """
    # Provide a list of geotess.util.Attribute tuples
    # No constraints on capitalization or spacing.
    attrib_names = "; ".join([attrib.name for attrib in attrib_tuples])
    attrib_units = "; ".join([attrib.unit for attrib in attrib_tuples])

    return attrib_names, attrib_units

def _layers_from_names_ids(gt_names, tess_ids):
    """Produce geotess.Layer tuples from GeoTess-style names and tess ids.
    """
    names = gt_names.split(';')

    return [Layer(name=name, tess_id=tess_id) for name, tess_id in zip(names, tess_ids)]

def _geotess_earth_shape(earth_shape, rconst):
    """Combine PyGeoTess earth shape and rconst values into GeoTess values.
    """
    if earth_shape == 'sphere' or not rconst:
        rc = ''
    else:
        rc = '_rconst'
    shp = earth_shape + rc

    return shp.upper()


class Model(object):
    """
    A class representing a 2D or 3D gridded Earth model.

    A GeoTess Model is comprised of 2D triangular tessellations of a unit
    sphere with 1D radial arrays of nodes associated with each vertex of
    the 2D tessellations. Variable spatial resolution in both geographic
    and radial dimensions is supported. 

    Parameters
    ----------
    gridfile : str
        Full path to an existing GeoTess grid file.  If it's an ascii grid,
        the extension should be ".ascii".  Because grid files are made using
        GeoTessBuilder and are required for a model, if gridfile is omitted,
        it is assumed that an empty Model class is desired, and the rest of the
        arguments are ignored.
    layers : list of geotess.util.Layer tuples
        Layer[0] is the layer name string
        Layer[1] is the layer tessellation id integer
    attributes : list of geotess.util.Attribute tuples
        Attribute[0] is the attribute name string
        Attribute[1] is the attribute unit string
    dtype : str {'double', 'float', 'long', 'int', 'shortint', 'byte'}
        Data type used to store attribute values.
    earth_shape : str {"sphere", "grs80", "wgs84"}
        Defines the ellipsoid that is to be used to convert between
        geocentric and geographic latitude and between depth and radius.
        The default is wgs84.

        * sphere - Geocentric and geographic latitudes are identical and
        conversion between depth and radius assume the Earth is a sphere
        with constant radius of 6371 km.
        * grs80 - Conversion between geographic and geocentric latitudes,
        and between depth and radius are performed using the parameters of
        the GRS80 ellipsoid.
        * wgs84 - Conversion between geographic and geocentric latitudes,
        and between depth and radius are performed using the parameters of
        the WGS84 ellipsoid.
        * iers2003 - Conversion between geographic and geocentric
        latitudes, and between depth and radius are performed using the
        parameters of the IERS2003 ellipsoid.
    rconst : bool
        If True, assume, for purposes of converting between depth and
        radius, that the radius of the earth is a constant equal
        to 6371 km.  Otherwise, assume that the radius decreases from
        equator to poles according to the specified earth_shape. 
    description : str, optional
        Plain-language description of the model.

    Attributes
    ----------
    layers : tuple of Layer tuples
    attributes : tuple of Attribute tuples

    References
    ----------
    http://www.sandia.gov/geotess

    """
    def __init__(self, gridfile=None, layers=None, attributes=None, dtype=None,
                 earth_shape="wgs84", rconst=False, description=None):

        self.layers = layers
        self.attributes = attributes

        if gridfile is None:
            self._model = None
            self.grid = None
        else:

            # Create GeoTessMetaData from inputs
            md = lib.GeoTessMetaData()
            if not description:
                description = ""
            md.setDescription(description)

            layer_names = _geotess_layer_names([layer.name for layer in layers])
            md.setLayerNames(layer_names)
            md.setLayerTessIds([int(layer.tess_id) for layer in layers])

            attrib_names, attrib_units = _geotess_attributes(attributes)
            md.setAttributes(attrib_names, attrib_units)

            earth_shape = _geotess_earth_shape(earth_shape, rconst)
            md.setEarthShape(earth_shape)

            md.setDataType(dtype.upper())
            md.setModelSoftwareVersion("PyGeoTess v{}".format(__version__))
            md.setModelGenerationDate(str(datetime.now()))

            # instantiate and store the low-level model instance
            self._model = lib.GeoTessModel(gridfile, md)

            self.grid = Grid.from_geotessgrid(self._model.getGrid())

    @classmethod
    def read(cls, modelfile):
        """
        Construct a Model instance from an existing model file.

        """
        m = cls()

        # load the model, add it as data on the instance
        model = lib.GeoTessModel()
        model.loadModel(modelfile)
        m._model = model

        # populate .attributes and .layers from the GeoTessMetaData
        md = model.getMetaData()

        attribute_names = md.getAttributeNamesString()
        attribute_units = md.getAttributeUnitsString()
        attributes = _attributes_from_strings(attribute_names, attribute_units)
        m.attributes = attributes

        layer_names = md.getLayerNamesString()
        layer_tess_ids = md.getLayerTessIds()
        layers = _layers_from_names_ids(layer_names, layer_tess_ids)
        m.layers = layers

        m.grid = Grid.from_geotessgrid(model.getGrid())

        return m

    def write(self, outfile):
        """
        Write the model to a file on disk.

        """
        self._model.writeModel(outfile)

    def vertices(self, layer=None, level=None, masked=False):
        """
        Get geographic coordinates of model vertices.

        Parameters
        ----------
        layer : int or str
            The integer index or layer name of the target layer.
        level : int
            The integer of the target tessellation level.
        masked : bool
            If False, don't return masked triangles.  Otherwise, return them all.
            Not yet implemented.

        Returns
        -------
        points : numpy.ndarray of floats (Nvert X Ndim)
            Geographic coordinates of points.
            For 2D models, vertices is Nvert X 2 (lon, lat).
            For 3D models, vertices is Nvert X 3 (lon, lat, radius).

        """
        try:
            # it's an integer layer index
            tessellation = self.layers[layer].tess_id
        except TypeError:
            # it's a string layer name
            tessellation = dict(self.layers)[layer]

        # XXX: finish this


    def points(self, layer=None, level=None):
        """
        Get geographic coordinates of model points.

        Parameters
        ----------
        layer : int or str
            The integer index or layer name of the target layer.
        level : int
            The integer of the target tessellation level.
        masked : bool
            If False, don't return masked triangles.  Otherwise, return them all.
            Not yet implemented.

        Returns
        -------
        points : numpy.ndarray of floats (Nvert X Ndim)
            Geographic coordinates of points.
            For 2D models, vertices is Nvert X 2 (lon, lat).
            For 3D models, vertices is Nvert X 3 (lon, lat, radius).

        """
        try:
            # it's an integer layer index
            tessellation = self.layers[layer].tess_id
        except TypeError:
            # it's a string layer name
            tessellation = dict(self.layers)[layer]

        # XXX: finish this

    def interpolate(self, domain, attribute, interpolant='linear'):
        pass

    def weights(self, domain, interpolant='linear'):
        pass

    def set_profile(self, vertex, layer, radii, attribute_values):
        """
        Set profile values for a vertex in a layer.

        """
        pass

    def __str__(self):
        return self._model.toString()




from glob import glob
import sys

from distutils.core import setup
from distutils.extension import Extension

import numpy as np

# see http://stackoverflow.com/a/4515279/745557
# and http://stackoverflow.com/a/18418524/745557

CPPFILES = glob('geotess/src/*.cc') # GeoTess c++ source (automatically finds .h files)
PYXFILES = ['geotess/src/libgeotess.pyx'] # hand-crafted Cython (automatically finds clibgeotess.pxd)
CYFILES = glob('geotess/src/libgeotess.cpp') # pre-cythonized c++ source files, in case cythonize fails

try:
    from Cython.Build import cythonize
    use_cython = True
except ImportError:
    use_cython = False

if use_cython:
    extensions = [Extension(name='geotess.libgeotess',
                  sources=CPPFILES+PYXFILES, language='c++',
                  include_dirs=[np.get_include()])]
    extensions = cythonize(extensions)
else:
    extensions = [Extension(name='geotess.libgeotess',
                  sources=CPPFILES+CYFILES, language='c++',
                  include_dirs=[np.get_include(), 'geotess/src'])]


setup(name = 'pygeotess',
      version = '0.2.1',
      description = 'GeoTess access from Python.',
      author = 'Jonathan K. MacCarthy',
      author_email = 'jkmacc@lanl.gov',
      packages = ['geotess'],
      py_modules = ['geotess.model', 'geotess.exc'],
      ext_modules = extensions,
      data_files = [ ('geotess/data', glob('geotess/src/GeoTessModels/*')) ],
      install_requires = [
          'numpy',
          ]
      )

